from typing import Tuple, List, Optional

import numpy as np

from .ILayerListener import ILayerListener
from ..Listenable import Listenable
from ..constants import Direction
from ..constants.CellValue import CellValue


class Layer(Listenable[ILayerListener]):
    def __init__(self, size: Tuple[int, int], defaultValue: CellValue):
        super().__init__()
        self.__size = size
        self.__defaultValue = defaultValue
        self.__cells = np.full([size[0] + 2, size[1] + 2], defaultValue, dtype=np.int32)

    @property
    def size(self) -> Tuple[int, int]:
        return self.__size

    @property
    def width(self) -> int:
        return self.__size[0]

    @property
    def height(self) -> int:
        return self.__size[1]

    @property
    def cells(self) -> np.ndarray:
        return self.__cells[1:-1, 1:-1]

    def __getitem__(self, coords: Tuple[int, int]) -> CellValue:
        return self.getValue(coords)

    def getValue(self, coords: Tuple[int, int], direction: Optional[Direction] = None) -> CellValue:
        x, y = coords[0], coords[1]
        w, h = self.__size
        assert 0 <= x < w, f"Invalid x={x}"
        assert 0 <= y < h, f"Invalid y={y}"
        if direction:
            if direction == Direction.LEFT:
                return CellValue(self.__cells[x, y + 1])
            if direction == Direction.RIGHT:
                return CellValue(self.__cells[x + 2, y + 1])
            if direction == Direction.TOP:
                return CellValue(self.__cells[x + 1, y])
            if direction == Direction.BOTTOM:
                return CellValue(self.__cells[x + 1, y + 2])
        return CellValue(self.__cells[x + 1, y + 1])

    def __setitem__(self, coords: Tuple[int, int], value: CellValue):
        self.setValue(coords, value)

    def setValue(self, coords: Tuple[int, int], value: CellValue):
        x, y = coords[0], coords[1]
        assert 0 <= x < self.__size[0], f"Invalid x={x}"
        assert 0 <= y < self.__size[1], f"Invalid y={y}"
        self.__cells[x + 1, y + 1] = value

    def fill(self, value: CellValue):
        self.__cells[1:-1, 1:-1] = value

    # Codes

    def getNeighbors4(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int]:
        x, y = cell
        return CellValue(self.__cells[x, y + 1]), \
               CellValue(self.__cells[x + 1, y]), \
               CellValue(self.__cells[x + 1, y + 2]), \
               CellValue(self.__cells[x + 2, y + 1])

    def getAllNeighbors4(self) -> np.ndarray:
        w, h = self.__size
        top = self.__cells[1:w + 1, 0:h]
        left = self.__cells[0:w, 1:h + 1]
        right = self.__cells[2:w + 2, 1:h + 1]
        bottom = self.__cells[1:w + 1, 2:h + 2]
        return np.stack((left, top, bottom, right), axis=2)

    def getAreaNeighbors4(self, cellsBox: Tuple[int, int, int, int]) -> np.ndarray:
        minX, maxX, minY, maxY = cellsBox
        top = self.__cells[minX + 1:maxX + 1, minY:maxY]
        left = self.__cells[minX:maxX, minY + 1:maxY + 1]
        right = self.__cells[minX + 2:maxX + 2, minY + 1:maxY + 1]
        bottom = self.__cells[minX + 1:maxX + 1, minY + 2:maxY + 2]
        return np.stack((left, top, bottom, right), axis=2)

    def getNeighbors8(self, cell: Tuple[int, int]) -> Tuple[int, int, int, int, int, int, int, int]:
        x, y = cell
        n = np.delete(self.__cells[x:x + 3, y:y + 3].flatten(), 4)
        return CellValue(n[0]), CellValue(n[1]), CellValue(n[2]), CellValue(n[3]), \
               CellValue(n[4]), CellValue(n[5]), CellValue(n[6]), CellValue(n[7])

    def getAreaNeighbors8(self, cellsBox: Tuple[int, int, int, int]) -> np.ndarray:
        minX, maxX, minY, maxY = cellsBox
        topLeft = self.__cells[minX:maxX, minY:maxY]
        top = self.__cells[minX + 1:maxX + 1, minY:maxY]
        topRight = self.__cells[minX + 2:maxX + 2, minY:maxY]
        left = self.__cells[minX:maxX, minY + 1:maxY + 1]
        right = self.__cells[minX + 2:maxX + 2, minY + 1:maxY + 1]
        bottomLeft = self.__cells[minX:maxX, minY + 2:maxY + 2]
        bottom = self.__cells[minX + 1:maxX + 1, minY + 2:maxY + 2]
        bottomRight = self.__cells[minX + 2:maxX + 2, minY + 2:maxY + 2]
        return np.stack((topLeft, left, bottomLeft, top,
                         bottom, topRight, right, bottomRight), axis=2)

    def getAllNeighbors8(self) -> np.ndarray:
        w, h = self.__size
        topLeft = self.__cells[0:w, 0:h]
        top = self.__cells[1:w + 1, 0:h]
        topRight = self.__cells[2:w + 2, 0:h]
        left = self.__cells[0:w, 1:h + 1]
        right = self.__cells[2:w + 2, 1:h + 1]
        bottomLeft = self.__cells[0:w, 2:h + 2]
        bottom = self.__cells[1:w + 1, 2:h + 2]
        bottomRight = self.__cells[2:w + 2, 2:h + 2]
        return np.stack((topLeft, left, bottomLeft, top,
                         bottom, topRight, right, bottomRight), axis=2)

    # Layer listener

    def notifyCellChanged(self, cell: Tuple[int, int]):
        for listener in self.listeners:
            listener.cellChanged(self, cell)
