from __future__ import annotations

from abc import abstractmethod
from typing import Tuple, Optional, Union

from pygame import Surface
from pygame.rect import Rect

from core import Listenable
from .IComponentListener import IComponentListener
from ..IUIEventHandler import IUIEventHandler
from ..Mouse import Mouse
from ..theme.Theme import Theme


class Component(Listenable[IComponentListener]):
    def __init__(self, theme: Theme,
                 size: Optional[Tuple[int, int]] = None):
        super().__init__()
        self.__theme = theme
        if size is None:
            self.__area = Rect(0, 0, theme.viewSize[0], theme.viewSize[1])
        else:
            self.__area = Rect(0, 0, size[0], size[1])

    @property
    def theme(self) -> Theme:
        return self.__theme

    def contains(self, pixel: Tuple[int, int]) -> bool:
        return self.__area.collidepoint(pixel) != 0

    @property
    def area(self) -> Rect:
        return self.__area.copy()

    @property
    def topLeft(self) -> Tuple[int, int]:
        return int(self.__area.left), int(self.__area.top)

    @property
    def topRight(self) -> Tuple[int, int]:
        return int(self.__area.right), int(self.__area.top)

    @property
    def bottomLeft(self) -> Tuple[int, int]:
        return int(self.__area.left), int(self.__area.bottom)

    @property
    def bottomRight(self) -> Tuple[int, int]:
        return int(self.__area.right), int(self.__area.bottom)

    @property
    def size(self) -> Tuple[int, int]:
        return int(self.__area.width), int(self.__area.height)

    def resize(self, size: Tuple[int, int]):
        self.__area.size = size
    
    # Anchors
    
    @staticmethod
    def relativeAnchor(anchor: str, size: Tuple[int, int]) -> Tuple[int, int]:
        if anchor == "topLeft":
            return 0, 0
        elif anchor == "top":
            return size[0] // 2, 0
        elif anchor == "topRight":
            return size[0], 0
        elif anchor == "left":
            return 0, size[1] // 2
        elif anchor == "center":
            return size[0] // 2, size[1] // 2
        elif anchor == "right":
            return size[0], size[1] // 2
        elif anchor == "bottomLeft":
            return 0, size[1]
        elif anchor == "bottom":
            return size[0] // 2, size[1]
        elif anchor == "bottomRight":
            return size[0], size[1]
        else:
            raise ValueError(f"Invalid anchor {anchor}")

    @staticmethod
    def innerBorderSize(anchor: str, borderSize: int) -> Tuple[int, int]:
        if anchor == "topLeft":
            return borderSize, borderSize
        elif anchor == "top":
            return 0, borderSize
        elif anchor == "topRight":
            return -borderSize, borderSize
        elif anchor == "left":
            return borderSize, 0
        elif anchor == "center":
            return 0, 0
        elif anchor == "right":
            return -borderSize, 0
        elif anchor == "bottomLeft":
            return borderSize, -borderSize
        elif anchor == "bottom":
            return 0, -borderSize
        elif anchor == "bottomRight":
            return -borderSize, -borderSize
        else:
            raise ValueError(f"Invalid anchor {anchor}")

    def moveRelativeTo(self,
                       anchor: str,
                       other: Optional[Component] = None,
                       otherAnchor: str = "center",
                       shiftX: Optional[int] = None,
                       shiftY: Optional[int] = None,
                       borderSize: Optional[int] = None):
        x, y = self.relativeAnchor(anchor, self.size)
        if other is None:
            otherX, otherY = (0, 0)
            otherWidth, otherHeight = self.theme.viewSize
        else:
            otherX, otherY = other.topLeft
            otherWidth, otherHeight = other.size
        selfX, selfY = self.relativeAnchor(otherAnchor, (otherWidth, otherHeight))
        if borderSize is None:
            borderSize = self.theme.frameBorderSize

        if anchor == otherAnchor:
            defaultShiftX, defaultShiftY = self.innerBorderSize(anchor, borderSize)
            if shiftX is None:
                shiftX = defaultShiftX
            if shiftY is None:
                shiftY = defaultShiftY
        else:
            if shiftX is None:
                if selfX <= 0:
                    shiftX = -borderSize
                elif selfX >= otherWidth:
                    shiftX = borderSize
                else:
                    shiftX = 0

            if shiftY is None:
                if selfY <= 0:
                    shiftY = -borderSize
                elif selfY >= otherHeight:
                    shiftY = borderSize
                else:
                    shiftY = 0

        self.moveTo((
            otherX + selfX - x + shiftX,
            otherY + selfY - y + shiftY
        ))

    def moveTo(self, topLeft: Tuple[int, int]):
        self.__area.update(topLeft, self.size)

    def shiftBy(self, shift: Tuple[int, int]):
        self.__area.move_ip(shift)

    # Component

    def dispose(self):
        pass

    def needRefresh(self) -> bool:
        return True

    @abstractmethod
    def render(self, surface: Surface):
        raise NotImplementedError()

    def findMouseFocus(self, mouse: Mouse) -> Optional[Component]:
        if not isinstance(self, IUIEventHandler):
            return None
        if not self.contains(mouse.pixel):
            return None
        return self

    # Component Listener

    def notifyWorldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        for listener in self.listeners:
            listener.worldCellClicked(cell, mouse)

    def notifyWorldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        for listener in self.listeners:
            listener.worldCellEntered(cell, mouse, dragging)

    def notifyMainBrushSelected(self, layerName: str, value: Union[int, str]):
        for listener in self.listeners:
            listener.mainBrushSelected(layerName, value)

    def notifySecondaryBrushSelected(self, layerName: str, value: Union[int, str]):
        for listener in self.listeners:
            listener.secondaryBrushSelected(layerName, value)

    def notifyViewChanged(self, view: Tuple[int, int]):
        for listener in self.listeners:
            listener.viewChanged(view)
