from typing import Tuple, Union

import pygame

from core.constants import CellValue
from core.logic import Logic
from core.state import World
from tools.vector import vectorMulI, vectorSubI, vectorClampI
from .GameMode import GameMode
from ..Mouse import Mouse
from ..component.IComponentListener import IComponentListener
from ..component.frame import ResourcesFrame
from ..component.frame.MinimapFrame import MinimapFrame
from ..component.frame.PaletteFrame import PaletteFrame
from ..component.world.WorldComponent import WorldComponent
from ..theme.Theme import Theme


class EditGameMode(GameMode, IComponentListener):

    def __init__(self, theme: Theme, world: World):
        super().__init__(theme)
        self.__world = world
        self.__logic = Logic(world)
        self.__font = theme.getFont("default")

        # Components
        self.__worldComponent = WorldComponent(theme, world)

        self.__resourcesFrame = ResourcesFrame(theme)
        self.__resourcesFrame.moveRelativeTo("topLeft", self, "topLeft")

        self.__paletteFrame = PaletteFrame(theme, world)
        self.__paletteFrame.moveRelativeTo("bottom", self, "bottom")

        self.__minimapFrame = MinimapFrame(theme, world)
        self.__minimapFrame.moveRelativeTo("bottomRight", self, "bottomRight")

        # Components order
        self.addComponent(self.__worldComponent)
        self.addComponent(self.__resourcesFrame)
        self.addComponent(self.__paletteFrame)
        self.addComponent(self.__minimapFrame)

        # Brush
        self.__mainBrushLayer = "ground"
        self.__mainBrushValue: Union[int, str] = CellValue.GROUND_EARTH
        self.__secondaryBrushLayer = "ground"
        self.__secondaryBrushValue: Union[int, str] = CellValue.GROUND_SEA

        # Listener
        self.__worldComponent.registerListener(self)  # worldCellClicked, ...
        self.__minimapFrame.registerListener(self)  # viewChanged
        self.__paletteFrame.registerListener(self)  # mainBrushSelected, ...

    @property
    def world(self) -> World:
        return self.__world

    @property
    def view(self) -> Tuple[int, int]:
        return self.__worldComponent.view

    # Game loop

    def dispose(self):
        super().dispose()
        self.__worldComponent.removeListener(self)
        self.__minimapFrame.removeListener(self)
        self.__paletteFrame.removeListener(self)

    def update(self):
        self.__logic.executeCommands()

    # UI Event handler

    def processInput(self) -> bool:
        # Update view using keyboard state
        keys = pygame.key.get_pressed()
        newViewX, newViewY = self.__worldComponent.view
        tileset = self.theme.getTileset("ground")
        tileWidth, tileHeight = tileset.tileSize
        speedX = tileWidth * 2
        speedY = tileHeight * 2
        if keys[pygame.K_RIGHT]:
            newViewX += speedX
        if keys[pygame.K_LEFT]:
            newViewX -= speedX
        if keys[pygame.K_UP]:
            newViewY -= speedY
        if keys[pygame.K_DOWN]:
            newViewY += speedY
        newView = (newViewX, newViewY)

        # Clamp new view
        worldSize = vectorMulI(self.__world.size, tileset.tileSize)
        maxView = vectorSubI(worldSize, self.theme.viewSize)
        newView = vectorClampI(newView, 0, maxView)

        # Update only if changes
        if newView != self.__worldComponent.view:
            self.viewChanged(newView)
            return True
        return False

    # Component listener

    def __updateCell(self, cell: Tuple[int, int], mouse: Mouse):
        if mouse.button1 or mouse.button2:
            brushLayer = self.__mainBrushLayer
            brushValue = self.__mainBrushValue
            fill = mouse.button2
        elif mouse.button3:
            brushLayer = self.__secondaryBrushLayer
            brushValue = self.__secondaryBrushValue
            fill = False
        else:
            return
        Command = self.__logic.getSetLayerValueCommand(brushLayer)
        command = Command(cell, brushValue, fill)
        self.__logic.addCommand(command)

    def worldCellClicked(self, cell: Tuple[int, int], mouse: Mouse):
        self.__updateCell(cell, mouse)

    def worldCellEntered(self, cell: Tuple[int, int], mouse: Mouse, dragging: bool):
        if dragging:
            self.__updateCell(cell, mouse)

    def mainBrushSelected(self, layerName: str, value: Union[int, str]):
        self.__mainBrushLayer = layerName
        self.__mainBrushValue = value

    def secondaryBrushSelected(self, layerName: str, value: Union[int, str]):
        self.__secondaryBrushLayer = layerName
        self.__secondaryBrushValue = value
